# Part of flinks
# (C) Martin Bays 2008
# Released under the terms of the GPLv3

import sys, os
from string import *

from time import sleep
import re

import curses
from curses.wrapper import wrapper
from curses.textpad import Textbox
import curses.ascii as ascii

from IndexedHypertext import IndexedHypertext

class QuitException(Exception):
    pass

class Browser:
    def __init__(self, scr, config):

	self.scr = scr
	self.config = config
	(self.maxy,self.maxx) = scr.getmaxyx()

	(
		self.wpm, self.initialUrl, self.showPoint, self.blankBetween,
		self.linkDisplayNum, self.linkDisplayEmphNum,
		self.trickyDisplayNum, self.maxCached, self.paraBlankSpace,
		self.sentenceBlankSpace
		) = map(config.options.get, [
		    'wpm', 'initialUrl', 'showPoint', 'blankBetween',
		    'linkDisplayNum', 'linkDisplayEmphNum',
		    'trickyDisplayNum', 'maxCached', 'paraBlankSpace',
		    'sentenceBlankSpace'
		    ])


	self.wordline = self.maxy/2
	self.statusline = self.wordline + self.linkDisplayNum + 1
	self.infoline = self.statusline + 1
	self.commandline = self.infoline + 1
	self.linkline = self.wordline - (self.linkDisplayNum + 1)
	self.trickyline = self.wordline + 2

	for i in range(1,8):
	    curses.init_pair(i, i, curses.COLOR_BLACK)

	self.scr.timeout(0)

	self.scr.leaveok(1)
	curses.curs_set(0)

	self.delay = 60.0/self.wpm
	self.paused = True

	self.lastSearch = None

	self.indexedHypertext = None

	self.cachedDocuments = []

	# history *ends* with most recent url
	# unHistory *ends* with *current* url
	self.history = []
	self.unHistory = []

	self.tricky = [""]*self.trickyDisplayNum

    def displayCentred(self, line, str, attr=0):
	self.scr.addstr(line, (self.maxx-len(str))/2, str, attr)
    def clearLine(self, line):
	self.scr.move(line,0)
	self.scr.clrtoeol()
    def displayCleanCentred(self, line, str, attr=0):
	self.clearLine(line)
	self.displayCentred(line, str, attr)
	self.scr.refresh()
    def displayStatus(self):
	self.clearLine(self.statusline)
	text = 'WPM: %d' % self.wpm
	if self.showPoint:
	    text += '; word: %d' % (self.indexedHypertext.atWord+1)
	    text += '; sentence: %d' % (self.indexedHypertext.atSentence+1)
	    text += '; paragraph: %d' % (self.indexedHypertext.atPara+1)
	    text += '; link: %d' % (self.indexedHypertext.lastLink+1)
	if self.paused: text += '; PAUSED'
	self.displayCentred(self.statusline, text)
    def displayLinkListEntry(self, linknum, emph=False):
	line = self.linkline + linknum%self.linkDisplayNum
	if linknum < 0:
	    self.clearLine(line)
	    return
	attr = self.attrOfLinknum(linknum)
	if emph:
	    attr |= curses.A_BOLD
	    emphstr = '*'
	else:
	    emphstr = ''
	link = self.indexedHypertext.links[linknum]
	str = '%s%s "%s": %s' % (
		emphstr, self.strOfLinknum(linknum),
		self.abbreviateStr(self.cleanWord(link["word"]), 10),
		self.abbreviateStr(link["url"])
		)
	self.clearLine(line)
	self.displayCentred(line, str, attr)
    def displayLinkList(self):
	for linknum in range(
		self.indexedHypertext.lastLink - self.linkDisplayNum,
		self.indexedHypertext.lastLink+1):
	    emphasize = (self.indexedHypertext.lastLink - linknum <
		    self.linkDisplayEmphNum)
	    self.displayLinkListEntry(linknum, emphasize)

    def attrOfLinknum(self, linknum):
	c = (linknum%self.linkDisplayNum)%5+1
	if c >= 4: c+=1 # don't use blue (hard to see against black)
	return curses.color_pair(c)

    def strOfLinknum(self, linknum):
	return '[%d]' % ((linknum%self.linkDisplayNum)+1)

    def displayTrickyList(self):
	for i in range(self.trickyDisplayNum):
	    self.clearLine(self.trickyline + i)
	    self.displayCentred(
		    self.trickyline + i,
		    self.tricky[i]
		    )

    def addTricky(self, word):
	for t in self.tricky:
	    if word == t:
		# already in the list; don't repeat
		return
	for i in range(self.trickyDisplayNum-1, 0, -1):
	    self.tricky[i] = self.tricky[i-1]
	self.tricky[0] = word
	self.displayTrickyList()

    def isTricky(self, word):
	return (
		bool(re.search('\d', word)) or bool(re.search('\w-\w', word))
		or len(word) > 15
		)

    def cleanWord(self, word):
	return strip(word, ' ,.;:?!(){}')

    def abbreviateStr(self, str, maxLen = None):
	if maxLen is None: maxLen = self.maxx/2
	if len(str) <= maxLen: return str
	return '%s...%s' % (str[:(maxLen/2)-3], str[-((maxLen/2)-3):])

    def refreshMoved(self):
	self.showCurrentWord()
	self.displayStatus()
	self.displayLinkList()
	self.displayTrickyList()

    def go(self, command):
	if command[:2] == 'g ':
	    url = "www.google.com/search?q=%s" % command[2:]
	elif command[:4] == 'wks ':
	    url = "www.google.com/search?q=site:en.wikipedia.org %s" % command[4:]
	elif command[:3] == 'wk ':
	    url = "en.wikipedia.org/wiki/%s" % command[3:]
	elif command[:3] == 'rss':
	    url = os.getenv('HOME')+"/.rawdog/out.html"
	else:
	    url = command
	self.gotoNewUrl(url)

    def gotoNewUrl(self, url):
	if self.indexedHypertext:
	    self.history.append(self.indexedHypertext.url)
	self.gotoUrl(url)
	# Turns out to be neater to have the current document in unHistory:
	self.unHistory = [url]
    def goHistory(self, num=1):
	try:
	    self.gotoUrl(self.history[-num])
	    for i in range(num):
		self.unHistory.append(self.history.pop())
	except IndexError:
	    return None
    def goUnHistory(self, num=1):
	try:
	    self.gotoUrl(self.unHistory[-num-1])
	    for i in range(num):
		self.history.append(self.unHistory.pop())
	except IndexError:
	    return None

    def gotoUrl(self, url):
	old = self.indexedHypertext
	cached = self.findCachedDocument(url)
	if cached:
	    self.indexedHypertext = cached
	else:
	    self.displayCleanCentred(self.infoline, self.abbreviateStr("Loading %s" % url))
	    self.scr.refresh()
	    self.indexedHypertext = IndexedHypertext(url)

	if self.indexedHypertext.isEmpty():
	    self.displayCleanCentred(self.infoline, "Failed to load url!")
	else:
	    self.displayCleanCentred(self.infoline, self.abbreviateStr("%s" % url))
	self.refreshMoved()

	if old: self.cacheDocument(old)

    def cacheDocument(self, doc):
	if self.findCachedDocument(doc.url):
	    return
	else:
	    self.cachedDocuments.append((doc.url, doc))
	    if len(self.cachedDocuments) > self.maxCached:
		self.cachedDocuments.pop(0)
    def findCachedDocument(self, url):
	for (u, doc) in self.cachedDocuments:
	    if u == url: return doc
	return None

    def showCurrentWord(self, word=None):
	if word is None:
	    word = self.indexedHypertext.currentWord()

	tricky = self.isTricky(word)
	linknum = self.indexedHypertext.currentLink()

	if linknum is None:
	    if tricky: attr = curses.A_BOLD
	    else: attr = 0
	    self.displayCleanCentred(self.wordline, word, attr)
	else:
	    self.displayCleanCentred(self.wordline, word,
		    self.attrOfLinknum(linknum) | curses.A_BOLD)

	    self.displayLinkListEntry(linknum, True)

	    if linknum >= self.linkDisplayEmphNum:
		self.displayLinkListEntry(linknum-self.linkDisplayEmphNum,
			False)

	if tricky:
	    self.addTricky(self.cleanWord(word))

	if self.showPoint: self.displayStatus()

    def mainLoop(self):
	self.go(self.initialUrl)
	self.displayStatus()

	blankSpace = 0

	while True:
	    if not self.paused:
		if blankSpace == 0:
		    try:
			word = self.indexedHypertext.readWord()
			if self.indexedHypertext.atParaStart():
			    blankSpace = self.paraBlankSpace
			elif self.indexedHypertext.atSentenceStart():
			    blankSpace = self.sentenceBlankSpace

		    except StopIteration:
			word = None
		else:
		    blankSpace -= 1
		    self.displayCleanCentred(self.wordline, "")

		if blankSpace == 0:
		    self.showCurrentWord(word)

	    c = None
	    escBound = self.config.keyBindings.has_key(27)
	    metaBit = False
	    while c != -1:
		c = self.scr.getch()

		if not escBound:
		    if c == 27:
			metaBit = True
		    elif metaBit:
			# set high bit
			c |= 0x80
			metaBit = False

		binding = self.config.keyBindings.get(c)
		if binding is not None:
		    getattr(self, 'doKey'+binding[0]).__call__(*binding[1])

	    if self.blankBetween and (not self.paused and
		    not self.indexedHypertext.atEnd()):
		sleep(self.delay/2)
		self.clearLine(self.wordline)
		self.scr.refresh()
		sleep(self.delay/2)
	    else:
		sleep(self.delay)


    bindableMethods = {
	    # <name> : [<min args>, <max args>]
	    'ChangeSpeed' : [1,1],
	    'SeekSentence' : [0,1],
	    'SeekParagraph' : [0,1],
	    'SeekWord' : [0,1],
	    'SeekLink' : [0,1],
	    'History' : [0,1],
	    'FollowLink' : [1,1],
	    'Go' : [0,0],
	    'GoCurrent' : [0,0],
	    'Search' : [0,1],
	    'SearchAgain' : [0,1],
	    'Pause' : [0,0],
	    'Refresh' : [0,0],
	    'Quit' : [0,0],
	    }

    def doKeyChangeSpeed(self, amount):
	self.wpm += amount
	if self.wpm < 30: self.wpm = 30
	self.delay = 60.0/self.wpm
	self.displayStatus()

    def doKeySeekSentence(self, n=1):
	self.indexedHypertext.seekSentenceRelative(n)
	self.refreshMoved()
    def doKeySeekParagraph(self, n=1):
	self.indexedHypertext.seekParaRelative(n)
	self.refreshMoved()
    def doKeySeekWord(self, n=1):
	self.indexedHypertext.seekWordRelative(n)
	self.refreshMoved()
    def doKeySeekLink(self, n=1):
	self.indexedHypertext.seekLinkRelative(n)
	self.refreshMoved()

    def doKeyFollowLink(self, n):
	n -= 1
	nl = self.indexedHypertext.lastLink % self.linkDisplayNum
	if n <= nl:
	    linknum = self.indexedHypertext.lastLink + (n-nl)
	else:
	    linknum = self.indexedHypertext.lastLink - self.linkDisplayNum + (n-nl) 
	url = self.indexedHypertext.links[linknum]["url"]
	self.gotoNewUrl(url)

    def readLine(self, prompt, initial=''):
	promptX = max(1, self.maxx/6 - len(prompt))
	boxX = promptX + len(prompt)
	boxLen = min(self.maxx - boxX - 1, 2*self.maxx/3)

	self.scr.addstr(self.commandline, promptX, prompt)
	self.scr.refresh()

	curses.curs_set(1)

	box = Textbox(curses.newwin(1, boxLen, self.commandline, boxX))

	for char in initial:
	    box.do_command(char)

	def editHook(char):
	    # implement ^U, which TextBox oddly lacks by default
	    if char == ord(ascii.ctrl('u')):
		(y,x) = box.win.getyx()
		box.win.move(y,0)
		box.win.clrtoeol()
	    else:
		return char

	read = box.edit(editHook)

	curses.curs_set(0)

	self.clearLine(self.commandline)

	return read

    def doKeyGo(self):
	command = self.readLine("Go: ")
	self.go(strip(command))

    def doKeyGoCurrent(self):
	command = self.readLine("Go: ", self.indexedHypertext.url)
	self.go(strip(command))

    def doKeySearch(self, dir=1):
	pattern = strip(self.readLine("Search: "))
	self.indexedHypertext.search(pattern, dir)
	self.lastSearch = {"pattern": pattern, "dir": dir}
	self.refreshMoved()
    def doKeySearchAgain(self, dir=1):
	if dir not in [1, -1]:
	    return
	if self.lastSearch:
	    self.indexedHypertext.search(self.lastSearch["pattern"],
		    self.lastSearch["dir"]*dir)
	    self.refreshMoved()

    def doKeyPause(self):
	self.paused = not self.paused
	self.displayStatus()

    def doKeyHistory(self, n=-1):
	if n < 0:
	    self.goHistory(-n)
	else:
	    self.goUnHistory(n)

    def doKeyRefresh(self):
	self.scr.redrawwin()

    def doKeyQuit(self):
	raise QuitException
