# Part of flinks
# (C) Martin Bays 2008
# Released under the terms of the GPLv3

import sys, os
from string import *

import re

from lynxDump import lynxDump

def placeInList(ref, list):
    if list == []: return -1
    for i,e in enumerate(list):
	if ref < e:
	    return i-1
    return i

def getItemCapped(list, i):
    if i < 0: i = 0
    if i >= len(list): i = len(list)-1
    return list[i]

STATE_WHITESPACE = 0
STATE_NEWLINE = 1
STATE_WORD = 2
STATE_LINKNUM = 3

class IndexedHypertext:
    def __init__(self, url):
	self.url = url
	dumped, linkUrls = lynxDump(url)

	self.links = [ {"url": url, "word": ''} for url in linkUrls ]

	self.words = []

	# index word position of start of each sentence, paragraph, link
	self.sentenceIndex = [0]
	self.paraIndex = [0]
	self.linkIndex = []

	state = STATE_WHITESPACE
	wordNum = -1
	currentWord = ''
	sentenceEnded = False
	paraEnded = False
	for c in dumped:
	    if state in [STATE_WHITESPACE,STATE_NEWLINE]:
		if c in "\n\r":
		    if state == STATE_NEWLINE:
			paraEnded = True
			sentenceEnded = True
		    state = STATE_NEWLINE
		elif c in whitespace:
		    state = STATE_WHITESPACE
		else:
		    state = STATE_WORD
		    wordacc = c
		    wordNum += 1
		    if sentenceEnded:
			self.sentenceIndex.append(wordNum)
			sentenceEnded = False
		    if paraEnded:
			self.paraIndex.append(wordNum)
			paraEnded = False

	    elif state == STATE_WORD:
		if c in whitespace:
		    skipBrackets = 0
		    while True:
			def regexp():
			    return ('(.*?)' +
				    '\[\d*\].*?' * skipBrackets +
				    '\[(\d*)\](.*)')


			m = re.match(regexp(), wordacc)
			if m:
			    wordstart, linknum, wordend = m.groups()
			    ln = int(linknum)
			    if (ln == len(self.linkIndex)+1 and 
				    ln <= len(self.links)):
				self.linkIndex.append(wordNum)
				wordacc = wordstart+wordend
				self.links[ln-1]["word"] = wordacc
			    else:
				skipBrackets += 1
			else: break
		    self.words.append(wordacc)
		    if c in "\n\r": state = STATE_NEWLINE
		    else: state = STATE_WHITESPACE
		else:
		    wordacc += c
		    sentenceEnded = False
		    if c in ".!?":
			sentenceEnded = True
	# end for

	if state == STATE_WORD: self.words.append(wordacc)

	self.iterator = self.iteratorFrom(0)
	self.atWord = -1 # Word to read next is self.atWord+1
	self.atSentence = 0
	self.atPara = 0
	if len(self.linkIndex) and self.linkIndex[0] == 0:
	    self.lastLink = 0
	else: self.lastLink = -1


    def iteratorFrom(self, fromWord):
	return self.words[fromWord:].__iter__()

    def currentLink(self):
	try: 
	    if self.atWord == self.linkIndex[self.lastLink]:
		return self.lastLink
	except IndexError: pass
	return None

    def readWord(self):
	word = self.iterator.next()

	self.atWord += 1
	try:
	    if self.atWord == self.sentenceIndex[self.atSentence+1]:
		self.atSentence += 1
	except IndexError: pass
	try:
	    if self.atWord == self.paraIndex[self.atPara+1]:
		self.atPara += 1
	except IndexError: pass
	try:
	    while self.atWord == self.linkIndex[self.lastLink+1]:
		self.lastLink += 1
	except IndexError: pass

	return word

    def currentWord(self):
	if self.atWord == -1:
	    return ''
	else:
	    try:
		return self.words[self.atWord]
	    except IndexError:
		return ''

    def isEmpty(self):
	return len(self.words) == 0
    def atEnd(self):
	return self.atWord == len(self.words)-1

    def seekWord(self, toWord):
	self.atWord = toWord
	self.atSentence = placeInList(toWord, self.sentenceIndex)
	self.atPara = placeInList(toWord, self.paraIndex)
	self.lastLink = placeInList(toWord, self.linkIndex)
	self.iterator = self.iteratorFrom(toWord+1)

    def seekWordRelative(self, rel):
	self.seekWord(max(0, min(len(self.words)-1, self.atWord+rel)))

    def seekSentenceRelative(self, rel):
	self.seekWord(getItemCapped(self.sentenceIndex,self.atSentence+rel))
    def seekParaRelative(self, rel):
	self.seekWord(getItemCapped(self.paraIndex,self.atPara+rel))
    def seekLinkRelative(self, rel):
	self.seekWord(getItemCapped(self.linkIndex,self.lastLink+rel))

    def atSentenceStart(self):
	return self.atWord == self.sentenceIndex[self.atSentence]
    def atParaStart(self):
	return self.atWord == self.paraIndex[self.atPara]

    def search(self, pattern, dir=1):

	patternWords = split(pattern)

	if len(patternWords) == 0:
	    return

	if dir == 1:
	    irange = range(self.atWord+1, len(self.words))
	else:
	    irange = range(self.atWord-1, -1, -1)

	for i in irange:
	    matched = 0
	    for pi in range(len(patternWords)):
		word = self.words[i+pi]
		pword = patternWords[pi]
		f = find(word, pword)
		if (f == -1 or
			(pi > 0 and f > 0) or
			(pi == 0 and len(patternWords) > 1 and
			    len(word) - f != len(pword) # doesn't match to end
			    )):
		    break
		else:
		    matched += 1
	    if matched == len(patternWords):
		self.seekWord(i)
		break
